# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
This module contains the helper functions for error handling.

The module contains three functions:
    - invalid_file_format() which should be used to validate the source file received against the expected schema and
      post any issues identified to the data feed dashboard.

    - post_errors() which is used to post errors (apart from those identified by the invalid_file_format() function) to
      the data feed dashboard.

    - validate_datetime() which checks whether the values of the datetime column(s) of the source file are valid. Any
      datetime errors identified by this function should be passed to the post_errors() function.

The validate_datetime() function can be used to validate the datetime column(s) of the source file. The output
`df_invalid_datetime` from this function should be passed to the post_errors() function. For example,

>>> import pandas as pd
>>> import switch_api as sw
>>> api_inputs = sw.initialize(user_id, api_project_id) # set api_project_id to the relevant portfolio and user_id to
                                                       # your own user identifier
>>> test_df = pd.DataFrame({'DateTime':['2021-06-01 00:00:00', '2021-06-01 00:15:00', '', '2021-06-01 00:45:00'],
... 'Value':[10, 20, 30, 40], 'device_id':['xyz', 'xyz', 'xyz', 'xyz']})
>>> df_invalid_datetime, df = validate_datetime(df=test_df, datetime_col=['DateTime'], dt_fmt='%Y-%m-%d %H:%M:%S')
>>> if df_invalid_datetime.shape[0] != 0:
...     sw.error_handlers.post_errors(api_inputs, df_invalid_datetime, error_type='DateTime')

"""

from .error_handlers import invalid_file_format, validate_datetime, post_errors

__all__ = ['validate_datetime', 'invalid_file_format', 'post_errors']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Data Ingestion into the Switch Automation
=========================================

Complete package for ingestion data into the Switch Automation Platform.
"""
__all__ = ['dataset', 'analytics', 'pipeline', 'error_handlers', 'integration', 'initialize', 'cache']

from .initialize import initialize
from . import dataset
from . import analytics
from . import pipeline
from . import error_handlers
# from . import platform_insights
from . import integration
from . import cache


# import logging
# switch_log = logging.getLogger(__name__).addHandler(logging.NullHandler())
# switch_log.setLevel(logging.DEBUG)
#
# _ch = logging.StreamHandler(stream= sys.stdout)  # creates the handler
# _ch.setLevel(logging.INFO)  # sets the handler info
# _ch.setFormatter(logging.Formatter(INFOFORMATTER))  # sets the handler formatting
#
# # adds the handler to the global variable: log
# log.addHandler(_ch)
# https://dev.to/joaomcteixeira/setting-up-python-logging-for-a-library-app-6ml

__version__ = "0.2.25"



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Import required functions
from setuptools import setup, find_packages

# Call setup function
setup(
    author="Switch Automation Pty Ltd.",
    description="A complete package for data ingestion into the Switch Automation Platform.",
    long_description=open('README.md', 'r').read() + '\n\n' + open('HISTORY.md', 'r').read(),
    long_description_content_type='text/markdown',
    license='MIT License',
    name="switch_api",
    version="0.2.25",
    packages=find_packages(include=["switch_api", "switch_api.*"]),
    install_requires=['pandas', 'requests', 'azure-storage-blob', 'pandera[io]==0.7.1', 'azure-servicebus', 'msal>=1.11.0'],
    python_requires=">=3.8.*",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Other Audience',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Natural Language :: English',
    ]
)



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining the Task abstract base class which is inherited by the following specific task types:

- IntegrationTask
- DiscoverableIntegrationTask
- AnalyticsTask
- LogicModuleTask
- QueueTask
- EventWorkOrderTask

Also includes the Automation class which contains helper functions.

---------------
IntegrationTask
---------------
Base class used to create integrations between the Switch Automation Platform and other platforms, low-level services,
or hardware.

Examples include:
    - Pulling readings or other types of data from REST APIs
    - Protocol Translators which ingest data sent to the platform via email, ftp or direct upload within platform.

---------------------------
DiscoverableIntegrationTask
---------------------------
Base class used to create integrations between the Switch Automation Platform and 3rd party APIs.

Similar to the IntegrationTask, but includes a secondary method `run_discovery()` which triggers discovery of available
points on the 3rd party API and upserts these records to the Switch Platform backend so that the records are available
in Build - Discovery & Selection UI.

Examples include:
    - Pulling readings or other types of data from REST APIs


-------------
AnalyticsTask
-------------
Base class used to create specific analytics functionality which may leverage existing data from the platform. Each
task may add value to, or supplement, this data and write it back.

Examples include:
    - Anomaly Detection
    - Leaky Pipes
    - Peer Tracking

---------------
LogicModuleTask
---------------
Base class that handles the running, reprocessing and scheduling of the legacy logic modules in a way which enables
integration with other platform functionality.

---------
QueueTask
---------
Base class used to create data pipelines that are fed via a queue.

----------
Automation
----------
This class contains the helper methods used to register, deploy, and test the created tasks. Additional helper
functions for retrieving details of existing tasks on the Switch Automation Platform are also included in this module.

"""

from .pipeline import (IntegrationTask, QueueTask, LogicModuleTask, AnalyticsTask, DiscoverableIntegrationTask,
                       EventWorkOrderTask, logger)
from .automation import Automation
from .definitions import (IntegrationDeviceConfigPropertyDefinition, IntegrationDeviceDefinition,
                          AnalyticsSettings, EventWorkOrderFieldDefinition, IntegrationSettings)

__all__ = ['IntegrationTask', 'QueueTask', 'LogicModuleTask', 'AnalyticsTask', 'DiscoverableIntegrationTask',
           'EventWorkOrderTask','Automation', 'logger', 'IntegrationDeviceConfigPropertyDefinition',
           'IntegrationDeviceDefinition', 'EventWorkOrderFieldDefinition',
           'AnalyticsSettings', 'IntegrationSettings']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import logging
import sys
from uuid import UUID
from msal.oauth2cli.oidc import decode_id_token
import requests
from ._credentials_store._credentials_store import SwitchCredentials, store_credentials
from ._credentials_store._credentials_store import read_credentials
from ._msal._custom_application import CustomPublicClientApplication
from .._utils._constants import AUTH_ENDPOINT_DEV, AUTH_ENDPOINT_PROD, SWITCH_ENVIRONMENT

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(stream=sys.stdout)
consoleHandler.setLevel(logging.INFO)

logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


_authority = {
  "dev": {
    "uitemplateid": "dev",
    "base_url": "https://switchdevb2c.b2clogin.com/switchdevb2c.onmicrosoft.com",
  },
  "prod": {
    "base_url": "https://switchautomation.b2clogin.com/switchautomation.onmicrosoft.com",
    "uitemplateid": "default"
  }
}


def _validate_credentials(creds: SwitchCredentials, version: str):
    """Validate given Credentials

    Ensures that credentials are up to date and synced.

    Parameters
    ----------
    creds
        SwitchCredentials to validate
    version
        The Credentials version

    Returns
    -------
    bool
        Returns the True when credentials are valid; otherwise returns False

    """

    if creds is None or creds is False or creds.user_id == '' or creds.user_id is None or creds.api_token is None:
        return False

    if creds.version != version:
        return False

    try:
        decode_id_token(creds.api_token)
        return True
    except Exception as ex:
        logger.info('Cached token expired. Requesting login.')
        return False


def get_switch_credentials(environment: SWITCH_ENVIRONMENT, portfolio_id):
    """Read Credentials

    Reads credentials from credential store

    Parameters
    ----------
    environment
        Either development or production environment
    portfolio_id
        Get credentials for the given portfolio id

    Returns
    -------
    SwitchCredentials
        Returns the SwitchCredentials namedtuple containing credentials to call Switch APIs

    """

    login_context = fetch_login_context(environment=environment, api_project_id=portfolio_id)[1]

    # Fetch from credentials store if it's already been cached
    if login_context is not None:
        datacenter = login_context['dataCenter']
        key_prefix = f"{environment}_{datacenter}" if environment == 'Development' else datacenter
        cur_credentials = read_credentials(key_prefix, login_context['portfolioId'], login_context['portfolioName'])
        if _validate_credentials(cur_credentials, login_context['version']):
            return cur_credentials
    else:
        raise Exception('Unable to obtain Login Context to authenticate with Switch Automation. '
                        'Please contact your administrator.')

    # Credentials do not exist, take user through browser auth journey
    config_env = 'dev' if environment == 'Development' else 'prod'
    authority = f"{_authority[config_env]['base_url']}/{login_context['policyName']}"
    app = CustomPublicClientApplication(login_context['applicationId'], authority=authority)

    uitemplateid = _authority[config_env]['uitemplateid']
    prompt = login_context['prompt'] if login_context['prompt'] != '' else None
    auth_response = app.acquire_token_interactive_custom(scopes=[], prompt=prompt, port=51796,
                                                         auth_params={
                                                             "uitemplateid": uitemplateid,
                                                             "portfolioId": portfolio_id,
                                                             "environment": environment
    })

    id_token = auth_response['id_token']

    api_base_url = AUTH_ENDPOINT_DEV if environment == 'Development' else AUTH_ENDPOINT_PROD
    _, profile = fetch_user_profile_details(api_base_url, id_token, environment, portfolio_id)

    if profile is None:
        raise Exception('Unable to obtain User Profile. Please try again or contact your administrator.')

    credentials = SwitchCredentials(login_context['version'], profile['email'],
                                    profile['userId'], login_context['portfolioId'],
                                    login_context['portfolioName'], id_token, profile['apiEndpoint'],
                                    environment, profile['subscriptionKey'])

    store_credentials(credentials, key_prefix=key_prefix)

    return credentials


def fetch_user_profile_details(api_base_url: str, token: str, environment: SWITCH_ENVIRONMENT, api_project_id: UUID):
    """
    Fetch User Profile details with respect to the given environment and portfolio

    Parameters
    ----------
    api_base_url
        User Profile base url
    token
        Switch token for authentication
    environment
        Either development or production environment
    api_project_id
        Get UserProfile with respect to this portfolio id

    Returns
    -------
    tuple[str, any]
        Returns the User Profile details containing information to call Switch APIs
    """

    payload = {}
    headers = {
        'Content-Type': 'application/json; charset=utf-8',
        'Authorization': f'bearer {token}'
    }

    url = f"{api_base_url}/user-profile?portfolioId={api_project_id}&switchEnvironment={environment}"

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, None
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, None
    return response_status, response.json()


def fetch_login_context(environment: SWITCH_ENVIRONMENT, api_project_id: UUID):
    """
    Fetches login context to assist SSO with Switch Platform

    Parameters
    ----------
    environment
        Either development or production environment
    api_project_id
        Get UserProfile with respect to this portfolio id

    Returns
    -------
    tuple[str, any]
        Returns the Login Context containing information to authenticate with Switch
    """

    base_url = AUTH_ENDPOINT_DEV if environment == 'Development' else AUTH_ENDPOINT_PROD

    payload = {}
    headers = {
        'Content-Type': 'application/json; charset=utf-8'
    }

    url = f"{base_url}/login-context?portfolioId={api_project_id}&switchEnvironment={environment}&type=PyPkg"

    response = requests.request("GET", url, data=payload, timeout=20, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, None
    elif len(response.text) == 0:
        logger.error('No data returned for this API call. %s', response.request.url)
        return response_status, None
    return response_status, response.json()



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
A module for interacting with the insights platform services, etc of the Switch Automation Platform.
"""

# from .platform_insights import get_current_insights_by_equipment

# __all__ = ['get_current_insights_by_equipment']
